use std::cmp::{max, min};

pub const LEVEL_WIDTH: usize = 50;
pub const LEVEL_HEIGHT: usize = 25;

#[derive(Copy, Clone, Debug, PartialEq)]
pub enum LevelElement {
    Wall,
    Floor,
    StairDown,
    StairUp,
    Unknown,
}

#[derive(Copy, Clone, Debug, PartialEq)]
pub enum RoomType {
    Start,
    End,
    StairUp,
    StairDown,
    BasicRoom,
    EmptyRoom,
}

#[derive(Copy, Clone, Debug, PartialEq)]
pub struct Level {
    pub(crate) structure: [[LevelElement; LEVEL_HEIGHT]; LEVEL_WIDTH],
    pub(crate) discovered: [[bool; LEVEL_HEIGHT]; LEVEL_WIDTH],
    pub(crate) start_x: usize,
    pub(crate) start_y: usize,
    pub(crate) end_x: usize,
    pub(crate) end_y: usize,
}

impl Level {
    pub fn new() -> Level {
        let s = [[LevelElement::Wall; LEVEL_HEIGHT]; LEVEL_WIDTH];
        Level {
            structure: s,
            discovered: [[false; LEVEL_HEIGHT]; LEVEL_WIDTH],
            start_x: 1,
            start_y: 1,
            end_x: 1,
            end_y: 1,
        }
    }
    pub fn get_element(&self, x: i16, y: i16) -> Option<LevelElement> {
        if x < 0 || y < 0 {
            return None;
        }
        if x >= LEVEL_WIDTH as i16 || y >= LEVEL_HEIGHT as i16 {
            return None;
        }
        if self.discovered[x as usize][y as usize] {
            return Some(self.structure[x as usize][y as usize]);
        }
        return Some(LevelElement::Unknown);
    }
    /// discover the area with in the level around the given position
    pub fn discover(&mut self, x: i16, y: i16) {
        for x_r in x - 1..=x + 1 {
            for y_r in y - 1..=y + 1 {
                let tx = min(max(x_r, 0), (LEVEL_WIDTH - 1) as i16) as usize;
                let ty = min(max(y_r, 0), (LEVEL_HEIGHT - 1) as i16) as usize;
                self.discovered[tx][ty] = true;
            }
        }
        for x_r in x..=x + 2 {
            let tx = min(max(x_r, 0), (LEVEL_WIDTH - 1) as i16) as usize;
            self.discovered[tx][y as usize] = true;
            if self.structure[tx][y as usize] == LevelElement::Wall {
                break;
            }
        }
        for y_r in y..=y + 2 {
            let ty = min(max(y_r, 0), (LEVEL_HEIGHT - 1) as i16) as usize;
            self.discovered[x as usize][ty] = true;
            if self.structure[x as usize][ty] == LevelElement::Wall {
                break;
            }
        }
        for x_r in (x-2..x).rev() {
            let tx = min(max(x_r, 0), (LEVEL_WIDTH - 1) as i16) as usize;
            self.discovered[tx][y as usize] = true;
            if self.structure[tx][y as usize] == LevelElement::Wall {
                break;
            }
        }
        for y_r in (y-2..y).rev() {
            let ty = min(max(y_r, 0), (LEVEL_HEIGHT - 1) as i16) as usize;
            self.discovered[x as usize][ty] = true;
            if self.structure[x as usize][ty] == LevelElement::Wall {
                break;
            }
        }
    }
}

#[test]
fn test_get_element() {
    let l = Level::new();
    assert_eq!(l.get_element(0, 0).unwrap(), LevelElement::Wall);
    assert_eq!(l.get_element(4, 4).unwrap(), LevelElement::StairDown);
    assert_eq!(l.get_element(49, 24).unwrap(), LevelElement::Wall);
    assert_eq!(l.get_element(50, 25), None);
}
